#include "peg_base.hh"

namespace MetaModule::PEG
{

void PEGBase::init_dig_inouts(void) {
	// DigIO::PingBut{mdrivlib::PinPull::Up};
	// DigIO::CycleBut{mdrivlib::PinPull::Up};
	// DigIO::PingJack{mdrivlib::PinPull::Down};
	// DigIO::CycleJack{mdrivlib::PinPull::Down};
	// DigIO::TrigJack{mdrivlib::PinPull::Down};
	// DigIO::EOJack{mdrivlib::PinPull::None, mdrivlib::PinSpeed::Medium, mdrivlib::PinOType::PushPull};
	// DigIO::ClockBusOut{mdrivlib::PinPull::None, mdrivlib::PinSpeed::Medium, mdrivlib::PinOType::PushPull};
	// DigIO::DebugOut{mdrivlib::PinPull::None, mdrivlib::PinSpeed::VeryHigh, mdrivlib::PinOType::PushPull};
}

}
