#pragma once
#include "util/interp_array.hh"
extern const InterpArray<float, 2048> sinTable;
extern const InterpArray<float, 2048> tanTable;
extern const InterpArray<float, 2048> exp5Table;
extern const InterpArray<float, 2048> exp10Table;
extern const InterpArray<float, 2048> pow9Table;
extern const InterpArray<float, 2048> logTable;
