#pragma once
#include "util/static_string.hh"
using ModuleTypeSlug = StaticString<63>;
using BrandTypeSlug = StaticString<63>;
using BrandModuleSlug = StaticString<127>;
using PatchName = StaticString<63>;
